@@ -40,11 +40,13 @@
 #include "lib/hci.h"
 #include "lib/mgmt.h"
 
+#include "src/shared/util.h"
 #include "mainloop.h"
 #include "display.h"
 #include "packet.h"
 #include "btsnoop.h"
 #include "hcidump.h"
+#include "ellisys.h"
 #include "control.h"
 
 static bool hcidump_fallback = false;
@@ -104,7 +106,8 @@ static void mgmt_controller_error(uint16_t len, const void *buf)
 
 static const char *settings_str[] = {
 	"powered", "connectable", "fast-connectable", "discoverable",
-	"pairable", "link-security", "ssp", "br/edr", "hs", "le"
+	"pairable", "link-security", "ssp", "br/edr", "hs", "le",
+	"advertising", "secure-conn"
 };
 
 static void mgmt_new_settings(uint16_t len, const void *buf)
@@ -222,7 +225,7 @@ static void mgmt_device_connected(uint16_t len, const void *buf)
 		return;
 	}
 
-	flags = btohl(ev->flags);
+	flags = le32_to_cpu(ev->flags);
 	ba2str(&ev->addr.bdaddr, str);
 
 	printf("@ Device Connected: %s (%d) flags 0x%4.4x\n",
@@ -380,7 +383,7 @@ static void mgmt_device_found(uint16_t len, const void *buf)
 		return;
 	}
 
-	flags = btohl(ev->flags);
+	flags = le32_to_cpu(ev->flags);
 	ba2str(&ev->addr.bdaddr, str);
 
 	printf("@ Device Found: %s (%d) rssi %d flags 0x%4.4x\n",
@@ -482,7 +485,7 @@ static void mgmt_passkey_notify(uint16_t len, const void *buf)
 
 	ba2str(&ev->addr.bdaddr, str);
 
-	passkey = btohl(ev->passkey);
+	passkey = le32_to_cpu(ev->passkey);
 
 	printf("@ Passkey Notify: %s (%d) passkey %06u entered %u\n",
 				str, ev->addr.type, passkey, ev->entered);
@@ -615,9 +618,9 @@ static void data_callback(int fd, uint32_t events, void *user_data)
 			}
 		}
 
-		opcode = btohs(hdr.opcode);
-		index  = btohs(hdr.index);
-		pktlen = btohs(hdr.len);
+		opcode = le16_to_cpu(hdr.opcode);
+		index  = le16_to_cpu(hdr.index);
+		pktlen = le16_to_cpu(hdr.len);
 
 		switch (data->channel) {
 		case HCI_CHANNEL_CONTROL:
@@ -625,7 +628,9 @@ static void data_callback(int fd, uint32_t events, void *user_data)
 			break;
 		case HCI_CHANNEL_MONITOR:
 			packet_monitor(tv, index, opcode, data->buf, pktlen);
-			btsnoop_write(tv, index, opcode, data->buf, pktlen);
+			btsnoop_write_hci(tv, index, opcode, data->buf, pktlen);
+			ellisys_inject_hci(tv, index, opcode,
+							data->buf, pktlen);
 			break;
 		}
 	}
@@ -709,11 +714,11 @@ static void client_callback(int fd, uint32_t events, void *user_data)
 
 	if (data->offset > MGMT_HDR_SIZE) {
 		struct mgmt_hdr *hdr = (struct mgmt_hdr *) data->buf;
-		uint16_t pktlen = btohs(hdr->len);
+		uint16_t pktlen = le16_to_cpu(hdr->len);
 
 		if (data->offset > pktlen + MGMT_HDR_SIZE) {
-			uint16_t opcode = btohs(hdr->opcode);
-			uint16_t index = btohs(hdr->index);
+			uint16_t opcode = le16_to_cpu(hdr->opcode);
+			uint16_t index = le16_to_cpu(hdr->index);
 
 			packet_monitor(NULL, index, opcode,
 					data->buf + MGMT_HDR_SIZE, pktlen);
@@ -806,22 +811,62 @@ void control_server(const char *path)
 	server_fd = fd;
 }
 
+void control_writer(const char *path)
+{
+	btsnoop_create(path, BTSNOOP_TYPE_EXTENDED_HCI);
+}
+
 void control_reader(const char *path)
 {
 	unsigned char buf[MAX_PACKET_SIZE];
-	uint16_t index, opcode, pktlen;
+	uint16_t pktlen;
+	uint32_t type;
 	struct timeval tv;
 
-	if (btsnoop_open(path) < 0)
+	if (btsnoop_open(path, &type) < 0)
 		return;
 
+	switch (type) {
+	case BTSNOOP_TYPE_HCI:
+	case BTSNOOP_TYPE_UART:
+	case BTSNOOP_TYPE_EXTENDED_PHY:
+		packet_del_filter(PACKET_FILTER_SHOW_INDEX);
+		break;
+
+	case BTSNOOP_TYPE_EXTENDED_HCI:
+		packet_add_filter(PACKET_FILTER_SHOW_INDEX);
+		break;
+	}
+
 	open_pager();
 
-	while (1) {
-		if (btsnoop_read(&tv, &index, &opcode, buf, &pktlen) < 0)
-			break;
+	switch (type) {
+	case BTSNOOP_TYPE_HCI:
+	case BTSNOOP_TYPE_UART:
+	case BTSNOOP_TYPE_EXTENDED_HCI:
+		while (1) {
+			uint16_t index, opcode;
 
-		packet_monitor(&tv, index, opcode, buf, pktlen);
+			if (btsnoop_read_hci(&tv, &index, &opcode,
+							buf, &pktlen) < 0)
+				break;
+
+			packet_monitor(&tv, index, opcode, buf, pktlen);
+			ellisys_inject_hci(&tv, index, opcode, buf, pktlen);
+		}
+		break;
+
+	case BTSNOOP_TYPE_EXTENDED_PHY:
+		while (1) {
+			uint16_t frequency;
+
+			if (btsnoop_read_phy(&tv, &frequency,
+							buf, &pktlen) < 0)
+				break;
+
+			packet_simulator(&tv, frequency, buf, pktlen);
+		}
+		break;
 	}
 
 	close_pager();
