@@ -27,13 +27,14 @@
 #include <errno.h>
 #include <stdbool.h>
 #include <glib.h>
-#include <bluetooth/uuid.h>
 #include <gdbus/gdbus.h>
 
+#include "lib/uuid.h"
 #include "plugin.h"
 #include "adapter.h"
 #include "device.h"
 #include "profile.h"
+#include "service.h"
 #include "dbus-common.h"
 #include "error.h"
 #include "attrib/gattrib.h"
@@ -144,7 +145,7 @@ static const char * const location_enum[] = {
 	"rear-dropout", "chainstay", "rear-wheel", "rear-hub"
 };
 
-static const gchar *location2str(uint8_t value)
+static const char *location2str(uint8_t value)
 {
 	if (value < G_N_ELEMENTS(location_enum))
 		return location_enum[value];
@@ -165,7 +166,7 @@ static int str2location(const char *location)
 	return -1;
 }
 
-static gint cmp_adapter(gconstpointer a, gconstpointer b)
+static int cmp_adapter(gconstpointer a, gconstpointer b)
 {
 	const struct csc_adapter *cadapter = a;
 	const struct btd_adapter *adapter = b;
@@ -176,7 +177,7 @@ static gint cmp_adapter(gconstpointer a, gconstpointer b)
 	return -1;
 }
 
-static gint cmp_device(gconstpointer a, gconstpointer b)
+static int cmp_device(gconstpointer a, gconstpointer b)
 {
 	const struct csc *csc = a;
 	const struct btd_device *dev = b;
@@ -187,7 +188,7 @@ static gint cmp_device(gconstpointer a, gconstpointer b)
 	return -1;
 }
 
-static gint cmp_watcher(gconstpointer a, gconstpointer b)
+static int cmp_watcher(gconstpointer a, gconstpointer b)
 {
 	const struct watcher *watcher = a;
 	const struct watcher *match = b;
@@ -505,7 +506,7 @@ static void discover_desc(struct csc *csc, struct gatt_char *c,
 	ch->csc = csc;
 	memcpy(ch->uuid, c->uuid, sizeof(c->uuid));
 
-	gatt_find_info(csc->attrib, start, end, discover_desc_cb, ch);
+	gatt_discover_char_desc(csc->attrib, start, end, discover_desc_cb, ch);
 }
 
 static void update_watcher(gpointer data, gpointer user_data)
@@ -513,7 +514,7 @@ static void update_watcher(gpointer data, gpointer user_data)
 	struct watcher *w = data;
 	struct measurement *m = user_data;
 	struct csc *csc = m->csc;
-	const gchar *path = device_get_path(csc->dev);
+	const char *path = device_get_path(csc->dev);
 	DBusMessageIter iter;
 	DBusMessageIter dict;
 	DBusMessage *msg;
@@ -758,7 +759,7 @@ done:
 		g_attrib_send(csc->attrib, 0, opdu, olen, NULL, NULL, NULL);
 }
 
-static void discover_char_cb(GSList *chars, guint8 status, gpointer user_data)
+static void discover_char_cb(uint8_t status, GSList *chars, void *user_data)
 {
 	struct csc *csc = user_data;
 	uint16_t feature_val_handle = 0;
@@ -963,8 +964,6 @@ static int csc_adapter_probe(struct btd_profile *p, struct btd_adapter *adapter)
 	cadapter = g_new0(struct csc_adapter, 1);
 	cadapter->adapter = adapter;
 
-	csc_adapters = g_slist_prepend(csc_adapters, cadapter);
-
 	if (!g_dbus_register_interface(btd_get_dbus_connection(),
 						adapter_get_path(adapter),
 						CYCLINGSPEED_MANAGER_INTERFACE,
@@ -977,6 +976,8 @@ static int csc_adapter_probe(struct btd_profile *p, struct btd_adapter *adapter)
 		return -EIO;
 	}
 
+	csc_adapters = g_slist_prepend(csc_adapters, cadapter);
+
 	return 0;
 }
 
@@ -1022,29 +1023,37 @@ static void property_set_location(const GDBusPropertyTable *property,
 	uint8_t att_val[2];
 	struct controlpoint_req *req;
 
-	if (csc->pending_req != NULL)
-		return g_dbus_pending_property_error(id,
+	if (csc->pending_req != NULL) {
+		g_dbus_pending_property_error(id,
 					ERROR_INTERFACE ".InProgress",
 					"Operation already in progress");
+		return;
+	}
 
-	if (!(csc->feature & MULTI_SENSOR_LOC_SUPPORT))
-		return g_dbus_pending_property_error(id,
+	if (!(csc->feature & MULTI_SENSOR_LOC_SUPPORT)) {
+		g_dbus_pending_property_error(id,
 					ERROR_INTERFACE ".NotSupported",
 					"Feature is not supported");
+		return;
+	}
 
-	if (dbus_message_iter_get_arg_type(iter) != DBUS_TYPE_STRING)
-		return g_dbus_pending_property_error(id,
+	if (dbus_message_iter_get_arg_type(iter) != DBUS_TYPE_STRING) {
+		g_dbus_pending_property_error(id,
 					ERROR_INTERFACE ".InvalidArguments",
 					"Invalid arguments in method call");
+		return;
+	}
 
 	dbus_message_iter_get_basic(iter, &loc);
 
 	loc_val = str2location(loc);
 
-	if (loc_val < 0)
-		return g_dbus_pending_property_error(id,
+	if (loc_val < 0) {
+		g_dbus_pending_property_error(id,
 					ERROR_INTERFACE ".InvalidArguments",
 					"Invalid arguments in method call");
+		return;
+	}
 
 	req = g_new(struct controlpoint_req, 1);
 	req->csc = csc;
@@ -1172,9 +1181,9 @@ static const GDBusMethodTable cyclingspeed_device_methods[] = {
 	{ }
 };
 
-static int csc_device_probe(struct btd_profile *p,
-				struct btd_device *device, GSList *uuids)
+static int csc_device_probe(struct btd_service *service)
 {
+	struct btd_device *device = btd_service_get_device(service);
 	struct btd_adapter *adapter;
 	struct csc_adapter *cadapter;
 	struct csc *csc;
@@ -1219,9 +1228,9 @@ static int csc_device_probe(struct btd_profile *p,
 	return 0;
 }
 
-static void csc_device_remove(struct btd_profile *p,
-						struct btd_device *device)
+static void csc_device_remove(struct btd_service *service)
 {
+	struct btd_device *device = btd_service_get_device(service);
 	struct btd_adapter *adapter;
 	struct csc_adapter *cadapter;
 	struct csc *csc;
@@ -1248,7 +1257,7 @@ static void csc_device_remove(struct btd_profile *p,
 
 static struct btd_profile cscp_profile = {
 	.name		= "Cycling Speed and Cadence GATT Driver",
-	.remote_uuids	= BTD_UUIDS(CYCLING_SC_UUID),
+	.remote_uuid	= CYCLING_SC_UUID,
 
 	.adapter_probe	= csc_adapter_probe,
 	.adapter_remove	= csc_adapter_remove,
